#include <SPI_MSTransfer.h>
#include "Stream.h"
#include <SPI.h>
#include <Wire.h>

#define SPI_TRANSACTION SPISettings(4000000, MSBFIRST, SPI_MODE0)
SPI_MSTransfer *_slave_pointer; 

SPI_MSTransfer::SPI_MSTransfer(const char *data, uint8_t cs, SPIClass *SPIWire) {
  if ( !strcmp(data,"Serial") ) { serial_port = 0; }
  else if ( !strcmp(data,"Serial1") ) { serial_port = 1; }
  else if ( !strcmp(data,"Serial2") ) { serial_port = 2; }
  else if ( !strcmp(data,"Serial3") ) { serial_port = 3; }
  else if ( !strcmp(data,"Serial4") ) { serial_port = 4; }
  else if ( !strcmp(data,"Serial5") ) { serial_port = 5; }
  else if ( !strcmp(data,"Serial6") ) { serial_port = 6; }
  else if ( !strcmp(data,"Wire") ) { wire_port = 0; }
  else if ( !strcmp(data,"Wire1") ) { wire_port = 1; }
  else if ( !strcmp(data,"Wire2") ) { wire_port = 2; }
  else if ( !strcmp(data,"Wire3") ) { wire_port = 3; }
  else if ( !strcmp(data,"SPI1") ) { remote_spi_port = 1; }
  else if ( !strcmp(data,"SPI2") ) { remote_spi_port = 2; }
  else if ( !strcmp(data,"EEPROM") ) { eeprom_support = 1; }
  else if ( !strcmp(data,"SERVO") ) { servo_support = 1; }
  else if ( !strcmp(data,"FASTLED" ) ) { fastled_support = 1; }
  chip_select = cs; spi_port = SPIWire; 
  ::pinMode(cs,OUTPUT); // make sure CS is OUTPUT before deassertion.
  ::digitalWrite(cs,HIGH); // deassert the CS way before SPI initializes :)
}



void SPI_MSTransfer::assert() {
  spi_port->beginTransaction(SPI_TRANSACTION); ::digitalWriteFast(chip_select,LOW);
}
void SPI_MSTransfer::deassert() { ::digitalWriteFast(chip_select,HIGH); spi_port->endTransaction(); }
bool SPI_MSTransfer::command_ack_response(uint16_t *data, uint32_t len) {
  uint8_t resend_count = 0; uint32_t timeout = millis();
  while ( spi_port->transfer16(0xFFA5) != 0xA506 ) {
    if ( millis() - timeout > 100 ) {
      if ( resend_count < 3 ) {
        resend_count++; deassert(); delay(10); assert();
        for ( uint32_t i = 0; i < data[1]; i++ ) { delayMicroseconds(1); spi_port->transfer16(data[i]); }
        timeout = millis();
      }
      else {
        deassert(); return 0;
      }
    }
  }
  return 1;
}
void SPI_MSTransfer::digitalWriteFast(uint8_t pin, bool state) { digitalWrite(pin, state); }
void SPI_MSTransfer::digitalWrite(uint8_t pin, bool state) {
  assert();
  uint16_t data[5], checksum = 0, data_pos = 0;
  data[data_pos] = 0x9766; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = 0x0000; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = ((uint16_t)(pin << 8) | state); checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = checksum;
  for ( uint32_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
  if ( !command_ack_response(data, sizeof(data) / 2) ) return;
  for ( uint8_t i = 0; i < 100; i++ ) {
    if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
      uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2;
      for ( uint32_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); } checksum = 0;
      for ( uint32_t i = 0; i < buf[1] - 1; i++ ) checksum ^= buf[i];
      if ( checksum == buf[buf[1] - 1] ) {
        deassert(); return;
      }
    }
  }
  deassert(); return;
}
bool SPI_MSTransfer::digitalReadFast(uint8_t pin) { return digitalRead(pin); }
bool SPI_MSTransfer::digitalRead(uint8_t pin) {
  assert();
  uint16_t data[5], checksum = 0, data_pos = 0;
  data[data_pos] = 0x9766; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = 0x0001; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = pin; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = checksum;
  for ( uint32_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
  if ( !command_ack_response(data, sizeof(data) / 2) ) return 0;
  for ( uint8_t i = 0; i < 100; i++ ) {
    if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
      uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2;
      for ( uint32_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); } checksum = 0;
      for ( uint32_t i = 0; i < buf[1] - 1; i++ ) checksum ^= buf[i];
      if ( checksum == buf[buf[1] - 1] ) {
        deassert(); return buf[5];
      }
    }
  }
  deassert(); return 0;
}
void SPI_MSTransfer::pinMode(uint8_t pin, uint8_t state) {
  assert();
  uint16_t data[5], checksum = 0, data_pos = 0;
  data[data_pos] = 0x9766; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = sizeof(data) / 2; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = 0x0002; checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = ((uint16_t)(pin << 8) | state); checksum ^= data[data_pos]; data_pos++;
  data[data_pos] = checksum;
  for ( uint32_t i = 0; i < data[1]; i++ ) spi_port->transfer16(data[i]);
  if ( !command_ack_response(data, sizeof(data) / 2) ) return;
  for ( uint8_t i = 0; i < 100; i++ ) {
    if ( spi_port->transfer16(0xFFFF) == 0xAA55 ) {
      uint16_t buf[spi_port->transfer16(0xFFFF)]; buf[0] = 0xAA55; buf[1] = sizeof(buf) / 2;
      for ( uint32_t i = 2; i < buf[1]; i++ ) { delayMicroseconds(1); buf[i] = spi_port->transfer16(0xFFFF); } checksum = 0;
      for ( uint32_t i = 0; i < buf[1] - 1; i++ ) checksum ^= buf[i];
      if ( checksum == buf[buf[1] - 1] ) {
        deassert(); return;
      }
    }
  }
  deassert(); return;
}



























































void SPI_MSTransfer::begin(uint32_t baudrate) { // set remote serial port and baud
}
void SPI_MSTransfer::begin() { // set remote I2C/SPI ports
}
uint16_t SPI_MSTransfer::analogRead(uint8_t pin) {
}
void SPI_MSTransfer::analogWrite(uint8_t pin, uint16_t value) {
}
void SPI_MSTransfer::analogReadResolution(uint8_t value) {
}
void SPI_MSTransfer::analogWriteResolution(uint8_t value) {
}
size_t SPI_MSTransfer::write(uint8_t val) {
}
int SPI_MSTransfer::read(void) {
}
int SPI_MSTransfer::available(void) {
}
int SPI_MSTransfer::peek() {
}
void SPI_MSTransfer::beginTransmission(uint8_t addr) {
}
void SPI_MSTransfer::endTransmission() {
}
void SPI_MSTransfer::requestFrom(uint8_t address, uint8_t bytes) {
}
uint8_t SPI_MSTransfer::transfer(uint8_t data) {
}
uint16_t SPI_MSTransfer::transfer16(uint16_t data) {
}
bool SPI_MSTransfer::online() {
}
uint16_t SPI_MSTransfer::events() { // keeps watchdog happy if not accessing slave... keep in loop.
}
void SPI_MSTransfer::software_reset() {
}
uint8_t SPI_MSTransfer::status_update() {
}
void SPI_MSTransfer::beginTransaction(uint32_t baudrate, uint8_t msblsb, uint8_t dataMode) { // set remote serial port and baud
}
void SPI_MSTransfer::endTransaction() { // end remote SPI transaction
}
int SPI_MSTransfer::read(int addr) {
}
uint16_t SPI_MSTransfer::length() {
}
size_t SPI_MSTransfer::write(int addr, uint8_t data) {
}
void SPI_MSTransfer::update(int addr, uint8_t data) {
}
uint32_t SPI_MSTransfer::crc() {
}
//void SPI_MSTransfer::show(uint8_t pin, CRGB *array, uint16_t array_length) { // update Fastled pixels
//}
size_t SPI_MSTransfer::print(const char *p) {
  while (*p) {
    char c = *p++;
    write(c);
  }
  return 0;
}
size_t SPI_MSTransfer::println(const char *p) {
  while (*p) {
    char c = *p++;
    write(c);
  }
  write('\n');
  return 0;
}
void SPI_MSTransfer::flush() {
}














































































SPI_MSTransfer::SPI_MSTransfer(const char *data, const char *mode) {
  if ( !strcmp(data,"SLAVE") && !strcmp(mode,"STANDALONE") ) {
    ::pinMode(13, OUTPUT);  // detach pin13 from SPI0 for led use.
    SIM_SCGC6 |= SIM_SCGC6_SPI0; // enable slave clock
    SPI0_MCR |= SPI_MCR_HALT | SPI_MCR_MDIS; // stop
    SPI0_CTAR0_SLAVE = SPI_CTAR_FMSZ(15) & SPI0_CTAR0_SLAVE & (~(SPI_CTAR_CPOL | SPI_CTAR_CPHA) | 0x00 << 25);
    SPI0_RSER = 0x00020000;
    CORE_PIN14_CONFIG = PORT_PCR_MUX(2);
    CORE_PIN11_CONFIG = PORT_PCR_DSE | PORT_PCR_MUX(2);
    CORE_PIN12_CONFIG = PORT_PCR_MUX(2);
    CORE_PIN2_CONFIG =  PORT_PCR_PS | PORT_PCR_MUX(2); // this uses pin 2 for the CS so Serial2 can be used instead.
    SPI0_MCR &= ~SPI_MCR_HALT & ~SPI_MCR_MDIS; // start
    NVIC_ENABLE_IRQ(IRQ_SPI0); // enable CS IRQ
    _slave_pointer = this;
  }
} 








void spi0_isr(void) {
  static uint16_t spi_data[DATA_BUFFER_MAX]; uint16_t buffer_pos = 0;
  while ( !(GPIOD_PDIR & 0x01) ) {
    if ( SPI0_SR & 0xF0 ) {
      spi_data[buffer_pos] = SPI0_POPR;
      if ( buffer_pos++ >= DATA_BUFFER_MAX ) break;
    }
    if ( spi_data[1] && buffer_pos >= spi_data[1] ) {
      _slave_pointer->SPI_MSTransfer::process_slave_data(spi_data, spi_data[1]); break;
    }
  }
  SPI0_SR |= SPI_SR_RFDF;
}
void SPI_MSTransfer::process_slave_data(uint16_t *data, uint32_t len) {
  uint16_t process_crc = 0; for ( uint32_t i = 0; i < len - 1; i++ ) process_crc ^= data[i];
    for ( uint16_t i = 0; i < len; i++ ) {
      Serial.print(data[i]); Serial.print(" ");
    } Serial.println();
  if ( data[len - 1] != process_crc ) return;
  while ( !(GPIOD_PDIR & 0x01) ) { // wait here until MASTER confirms ACK receipt
    if ( SPI0_SR & 0xF0 ) {
      SPI0_PUSHR_SLAVE = 0xA506; if ( SPI0_POPR == 0xFFA5 ) break;
    }
  }
  switch ( data[0] ) {
    case 0x9766: {
        switch ( data[2] ) {
          case 0x0000: {
              ::digitalWriteFast(data[3] >> 8, data[3]);
              uint16_t checksum = 0, buf_pos = 0, buf[] = { 0xAA55, 0, data[0], data[2], data[3], 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint32_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) checksum ^= buf[i]; buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++; if ( buf_pos >= sizeof(buf) / 2) buf_pos = 0;
                }
              }
              return;
            }
          case 0x0001: {
              uint16_t checksum = 0, buf_pos = 0, buf[] = { 0xAA55, 0, data[0], data[2], data[3], ::digitalReadFast(data[3]), 0 }; buf[1] = sizeof(buf) / 2;
              for ( uint32_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) checksum ^= buf[i]; buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++; if ( buf_pos >= sizeof(buf) / 2) buf_pos = 0;
                }
              }
              return;
            }
          case 0x0002: {
              ::pinMode(data[3] >> 8, data[3]);
              uint16_t checksum = 0, buf_pos = 0, buf[] = { 0xAA55, 0, data[0] }; buf[1] = sizeof(buf) / 2;
              for ( uint32_t i = 0; i < (sizeof(buf) / 2) - 1; i++ ) checksum ^= buf[i]; buf[(sizeof(buf) / 2) - 1] = checksum;
              while ( !(GPIOD_PDIR & 0x01) ) {
                if ( SPI0_SR & 0xF0 ) {
                  SPI0_PUSHR_SLAVE = buf[buf_pos]; SPI0_POPR; buf_pos++; if ( buf_pos >= sizeof(buf) / 2) buf_pos = 0;
                }
              }
              return;
            }



        }
        break;
      }




  }

}