#ifndef _SPI_MSTransfer_H_
#define _SPI_MSTransfer_H_

#include "Stream.h"
#include <SPI.h>
#include <Wire.h>
#include <FastLED.h>

#define DATA_BUFFER_MAX 1024

class SPI_MSTransfer : public Stream {

  public:
    SPI_MSTransfer(const char *data, uint8_t addr, SPIClass *SPIWire);
    SPI_MSTransfer(const char *data, const char *mode);

    virtual void      digitalWrite(uint8_t pin, bool state);
    virtual void      digitalWriteFast(uint8_t pin, bool state);
    virtual bool      digitalRead(uint8_t pin);
    virtual bool      digitalReadFast(uint8_t pin);
    virtual void      pinMode(uint8_t pin, uint8_t state);
    virtual void      process_slave_data(uint16_t *data, uint32_t len);


    virtual void      beginTransaction(uint32_t baudrate, uint8_t msblsb, uint8_t dataMode); // SPI
    virtual void      endTransaction(); // SPI 
    virtual void      begin(uint32_t baudrate); // UART
    virtual void      begin(); // I2C/SPI
    virtual void      beginTransmission(uint8_t addr); // I2C
    virtual void      endTransmission(); // I2C
    virtual void      requestFrom(uint8_t address, uint8_t bytes); // I2C
    virtual uint8_t   transfer(uint8_t data); // SPI 8bit
    virtual uint16_t  transfer16(uint16_t data); // SPI 16bit
    virtual uint16_t  analogRead(uint8_t pin); // 16bit
    virtual void      analogReadResolution(uint8_t value);
    virtual void      analogWrite(uint8_t pin, uint16_t value); // 16bit
    virtual void      analogWriteResolution(uint8_t value);
    virtual int       read(); // I2C/UART
    virtual int       available(); // I2C/UART
    virtual size_t    write(uint8_t val); // I2C/UART
    virtual int       read(int addr); // EEPROM
    virtual size_t    write(int addr, uint8_t data); // EEPROM/PWMSERVO
    virtual void      update(int addr, uint8_t data); // EEPROM
    virtual uint16_t  length(); // EEPROM
    virtual uint32_t  crc(); // EEPROM
    virtual int       peek(); // I2C/UART
    //virtual void      show(uint8_t pin, CRGB *array, uint16_t array_length); // Fastled
    virtual size_t    print(const char *p);
    virtual size_t    println(const char *p);
    virtual void      flush();
    virtual void      assert();
    virtual void      deassert();
    virtual void      software_reset();
    virtual bool      online();
    virtual uint16_t  events();
  private:
    virtual bool       command_ack_response(uint16_t *data, uint32_t len);
    virtual  uint8_t   status_update();
    volatile uint8_t   chip_select = -1;
    volatile uint8_t   wire_port = -1;
    volatile uint8_t   eeprom_support = -1;
    volatile uint8_t   spi_support = -1;
    volatile uint8_t   fastled_support = -1;
    volatile uint8_t   servo_support = -1;
    volatile uint8_t   serial_port = -1;
    volatile uint8_t   remote_spi_port = -1;
    SPIClass           *spi_port;
};
#endif